//
//  resummationSCET.cc
//  ResHiggs
//
//  Created by Luca Rottoli on 29/10/2014.
//

#include <cmath>
#include "resummationSCET.hh"
#include "chebyshev.hh"
#include "QCD_EW.hh"
#include "math/special_functions.hh"
#include "math/mellin_Dk.hh"
#include "math/integration.hh"
#include "math/struct_definitions.hh"
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include <cuba.h>

#include "../../ggHiggs/src/NLO_finite-mt/bdv.hh"
#include "../../ggHiggs/src/NNLO_finite-mt/nnlo_finite-mt.hh"

#include "LHAPDF/LHAPDF.h"

using namespace std;


namespace resumm
{

  /**
    * SCETResummation destructor
    */
  SCETResummation::~SCETResummation()
  {
    return;
  }

  /**
    * Resummation
    */
  double SCETResummation::res_xs(double m, double sqrt_s,
                                 double muF_m_ratio, double muH_m_ratio,
                                 log_order LOGo, order match_order, double *asinput)
  {
    doRapidityDistr = false;
    
    tau  = m*m/sqrt_s/sqrt_s;
    M    = m;
    muF  = m*muF_m_ratio;
    muR  = muF;
    fmuh = m*muH_m_ratio;
    fmut = fmuh*fmutmuhrat; // if EFT, then Pi^2 res is only on CS

    if(fPi==true) fmuh *= -I;
    if(!mtEFT) fmut = fmuh; // in the exact theory, mut does not exist: must be equal to muh

    if(!musset) {
      double musI  = GetmusI (M,sqrt_s);
      double musII = GetmusII(M,sqrt_s);
      fmus= 1./2 * musI + (1./2) * musII;
    }

    if(prescr==MP00 && APimpr>0) {
      cout << "Do not use MP00 with APi, i>0" << endl;
    }

    LL_order = LOGo;
    matching_order = match_order;
    if(int(LL_order) < int(match_order)) {
      cout << "WARNING! Currently you can't match this logarithmic order " << LL_order << " to fixed order " << match_order << ". Use a smaller match_order." << endl;
      exit(0);
    }
    if(match_order==NNNLO) {
      cout << "WARNING! Matching to NNNLO not yet implemented." << endl;
      exit(0);
    }

    bool redo_cheb = !(last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommonSCET(redo_cheb);
    Chebyshev cheb(this);
    
    expDconst = true;
	
    double result = 0.;
    
    MP_slope = 1.;
    MP_reN0 = MPbeta + 3.;
    double NL = Nlandau();
    if(MP_reN0 >= NL)
      {
	MP_reN0 = ( MPbeta + NL )/2.;
	if(MP_reN0 >= NL)
	  {
	    cout << "ERROR: Landau pole N_L = " << NL
		 << " to the left of the rightmost singularity: " << MPbeta << endl;
	    exit(-13);
	  }
      }
    MPzmin=tau;
    if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
    result = KFactor()/luminosity_in_tau;
    //cout << LuminosityN(2) << endl;
    //cout << Subtraction(2) << endl;
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muH_m_ratio;
    last_M      = M;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = mtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }

  /**
   * Resummation
   */
  double SCETResummation::sigma_contribution(double m, double sqrt_s,
					     double muF_m_ratio, double muH_m_ratio,
					     log_order LOGo, order match_order, double mus, double *asinput)
  {
    doRapidityDistr = false;
	
    tau = m*m/sqrt_s/sqrt_s;
    M   = m;
    muF = m*muF_m_ratio;
    muR  = muF;
    fmuh = m*muH_m_ratio;
    fmut = fmuh*fmutmuhrat; // if EFT, then Pi^2 res is only on CS

    if(fPi==true) fmuh *= -I;
    if(!mtEFT) fmut = fmuh; // in the exact theory, mut does not exist: must be equal to muh

    fmus = mus;
	
    LL_order = LOGo;
    matching_order = match_order;
	
	
    bool redo_cheb = !(last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommonSCET(redo_cheb);
    Chebyshev cheb(this);
	
    expDconst = true;
    
    double result = 0.;
	
    MP_slope = 1.;
    MP_reN0 = MPbeta + 3.;
    double NL = Nlandau();
    if(MP_reN0 >= NL)
      {
	MP_reN0 = ( MPbeta + NL )/2.;
	if(MP_reN0 >= NL)
	  {
	    cout << "ERROR: Landau pole N_L = " << NL
		 << " to the left of the rightmost singularity: " << MPbeta << endl;
	    exit(-13);
	  }
      }
    MPzmin=tau;
    if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
    result = StildeFactor(fmus)/luminosity_in_tau;
	
    last_muFrat = muF_m_ratio;
    last_muRrat = muH_m_ratio;
    last_M      = M;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = mtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit   = false;
    //
    return result;
  }
  
  
  double SCETResummation::res_xs_rap(double m, double Y, double sqrt_s,
				    double muF_m_ratio, double muH_m_ratio,
				    log_order LOGo, order match_order, double *asinput)
  {
    return 0;
  }

  double SCETResummation::GetmusI(double m, double sqrts) {
    double a1, b1, c1;
    a1 = 0.999;
    b1 = 0.0721;
    c1 = 32.47;
    if(APimpr==AP1) {
      a1 = 2.36889;
      b1 = 3.92861;
      c1 = 0.705348;
    }
    double t = m*m/sqrts/sqrts;
    double musI = m * (1.-t) / pow((a1 + b1*pow(t,0.25)),c1);
    return musI;
  }
  double SCETResummation::GetmusII(double m, double sqrts) {
    double a2, b2, c2;
    a2 = 1.098;
    b2 = 0.344;
    c2 = 7.184;
    if(APimpr==AP1) {
      a2 = 1.57468;
      b2 = 0.943819;
      c2 = 2.72502;
    }
    double t = m*m/sqrts/sqrts;
    double musII = m * (1.-t) / pow((a2 + b2*pow(t,0.25)),c2);
    return musII;
  }
  
  /**
    * Variable initialisation
    */
  void SCETResummation::ComputeCommonSCET(bool redo_cheb)
  {
    ComputeCommon(redo_cheb);
    InitResummCoeffsSCET();
  }
  
  /**
    * KFactor
    */
  double SCETResummation::KFactor()
  {
    InitAlphas();
    H();
    U();
    Eta();
    return SCETMinimalPrescription();
  }
  
  /**
    * StildeFactor
    */
  double SCETResummation::StildeFactor(double mus)
  {
    InitAlphas();
    Eta();
    return StildeMinimalPrescription();
  }
  
  
  dcomplex SCETResummation::FO_1_log_PL(dcomplex N) {
    using namespace QCD;
    double lf = -2.*log(muF/M);
    return 4.*CA*mellD(1,N) + 2*CA*mellD(0,N)*lf + 2.*CA * ZETA2;
  }
  dcomplex SCETResummation::FO_2_log_PL(dcomplex N) {
    using namespace QCD;
    double lf = -2.*log(muF/M);
    dcomplex value;
    value = ( 72. * mellD(3,N)
	      + (2.*nf - 33.) * mellD(2,N)
	      + (67. - 90.*ZETA2 - 10./3.*nf) * mellD(1,N)
	      + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + nf*(14./9.-2.*ZETA2)) * mellD(0,N)
	      + 837./16. + 67./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3 + nf  * ( 5./6.*ZETA3 - 5./3.*ZETA2 - 247./36. )
	      );
    value += ( 108*lf * mellD(2,N)
	       + (36.*lf*lf - (33-2*nf)*lf) * mellD(1,N)
	       + (-(33-2*nf)/4*lf*lf + (67./2.-5./3.*nf-5.*CA*CA*ZETA2)*lf) * mellD(0,N)
	       - 18.*ZETA2*lf*lf + (-3./4. + 171./2.*ZETA3 + 1./4.*nf - (33.-2.*nf)/2.*ZETA2)*lf
	       );
    return value;
  }
  dcomplex SCETResummation::FO_3_log_PL(dcomplex N) {
    using namespace QCD;
    //double lf = -2.*log(muF/M);
    double CA3 = CA*CA*CA;
    dcomplex value;
    value = ( 8*CA3 * mellD(5,N)
	      + (-110./9.*CA3 + 20/9.*CA*CA*nf) * mellD(4,N)
	      + (CA3*(925./27. - 56*ZETA2) - 164./27.*CA*CA*nf + 4./27.*CA*nf*nf) * mellD(3,N)
	      + (CA3 * (-1051./27. + 187./3*ZETA2 + 181*ZETA3)
		 + CA*CA*nf * (457./2./27. - 34./3.*ZETA2)
		 + 0.5*CA*CF*nf - 10./27.*CA*nf*nf ) * mellD(2,N)
	      + (CA3 * (30569./8./81. - 152./3.*ZETA2 - 352./3.*ZETA3 - 154./5.*ZETA2*ZETA2)
		 + CA*CA*nf * (-4211./4./81. + 94./9.*ZETA2 + 46./3.*ZETA3)
		 - CA*CF*nf * (63./8.-6.*ZETA3)
		 + CA*nf*nf * (25./81.-4./9.*ZETA2) ) * mellD(1,N)
	      + (CA3 * (-297029./32./729. + 8563./4./81.*ZETA2 + 8941./4./27.*ZETA3 + 253./4./15.*ZETA2*ZETA2
			- 725./2./3.*ZETA2*ZETA3 + 186.*ZETA5)
		 + CA*CA*nf * (31313./16./729. - 2173./4./81.*ZETA2 - 475./4./9.*ZETA3 - 17./2./15*ZETA2*ZETA2)
		 + CA*CF*nf * (1711./32./27. - 0.5*ZETA2 - 19./2./9.*ZETA3 - 1./5.*ZETA2*ZETA2)
		 - CA*nf*nf * (58./729. - 10./27.*ZETA2 - 5./27.*ZETA3) ) * mellD(0,N)
	      + 1740.02 - 131.928*nf + 1.75719*nf*nf
	      );
    /*
    value += ( (16.*CA*CA*CA*lf*lf + 0*lf) * mellD(3,N)
	    + (4.*CA*CA*CA*lf*lf*lf - 18.*CA*CA*b0*lf*lf + 12.*CA*A1*lf) * mellD(2,N)
	    + (-4.*CA*CA*b0*lf*lf*lf + (8.*CA*A1+4.*CA*b0*b0+4.*CA*CA*W1-32.*CA*CA*CA*ZETA2)*lf*lf + (8*CA*B1-12*CA*b1)*lf) * mellD(1,N)
	    + ((2.*b0*b0-4.*CA*CA*CA*ZETA2)*lf*lf*lf + (CA*(4.*B1-5.*b1)-3.*CA*b0*W1-2.*A1*b0+6.*CA*CA*b0*ZETA2+32.*CA*CA*CA*ZETA3)*lf*lf
	       + (2.*A2+2.*W1*A1-4.*CA*A1*ZETA2)*lf) * mellD(0,N) );
  // adding contributions from interference with C2
    value += lf*(2.*CA*( 72. * D0D3(z,D)
		      + (2.*nf - 33.) * D0D2(z,D)
		      + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*nf) * D0D1(z,D)
		      + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + nf*(14./9.-2.*ZETA2)) * 2*D(1,z)
		      + (_cd2 + W2 + _cd1*W1 ) * D(0,z)
		      )
	      -2.*b0*( 72. * D(3,z)
		       + (2.*nf - 33.) * D(2,z)
		       + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*nf) * D(1,z)
		       + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + nf*(14./9.-2.*ZETA2)) * D(0,z)
		       ) );
  c += (72.*ZETA3+18.*b0*ZETA2)*lf*lf*lf
    + (-4.*CA*A1*ZETA2-2.*b0*(B1-b1+6.*CA*CA*ZETA3)+b0*b0*(2.*CA*ZETA2+W1)-2.*CA*CA*W1*ZETA2+4.*CA*CA*CA*(ZETA2*ZETA2-6.*ZETA4))*lf*lf
    + (4.*CA*B1*ZETA2-6.*CA*b1*ZETA2-3.*W1*b1+2.*B2-2.*b2+2.*W1*B1+8.*CA*A1*ZETA3)*lf;
  // adding contributions from interference with C2
  c += lf*(2*CA*( 72. * 6*ZETA5
		  + (2.*nf - 33.) * (-2*ZETA4)
		  + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*nf) * ZETA3
		  + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + nf*(14./9.-2.*ZETA2)) * (-ZETA2)
		  )
	   -  2*b0* (_cd2 + W2 + _cd1*W1 ) );
	   double D0D1(double z, double D(int,double)) {
	   return 3./2.*D(2,z) - ZETA2*D(0,z);
	   }
	   double D0D2(double z, double D(int,double)) {
	   return 4./3.*D(3,z) - 2.*ZETA2*D(1,z) + 2.*ZETA3*D(0,z);
	   }
	   double D0D3(double z, double D(int,double)) {
	   return 5./4.*D(4,z) - 3.*ZETA2*D(2,z) + 6.*ZETA3*D(1,z) - 6.*ZETA4*D(0,z);
	   }
 */
    return value;
  }
  
  /**
    * Fixed Order matching
    */
  dcomplex SCETResummation::Subtraction(dcomplex N)
  {
    using namespace QCD;	
    dcomplex value = 1.;
    dcomplex fSub1, fSub2, fSub3;
    //
    if (matching_order > 0) {
      fSub1 = fasmuf/M_PI*(FO_1_log_PL(N)+W1);
      value += fSub1;
    }
    if (matching_order > 1) {
      fSub2 = pow(fasmuf/M_PI,2)*(FO_2_log_PL(N) + W1*FO_1_log_PL(N) + W2 )  ;
      value += fSub2;
    }
    if (matching_order > 2) {
      fSub3 = pow(fasmuf/M_PI,3)*(FO_3_log_PL(N) + W2*FO_1_log_PL(N) + W1*FO_2_log_PL(N) + W3 )  ;
      value += fSub3;
    }
    return value;
  }
  
  /**
	* One Loop Soft-Function Contribution
	*/
  dcomplex SCETResummation::StildeOneLoop(dcomplex N)
  {
	using namespace QCD;
	dcomplex gammaplus;
	dcomplex gammaminus;
	dcomplex funeta;
	dcomplex d0;
	dcomplex value;
	fEta = 0;//fEta1;
	gammaplus = Gamma(N+fEta);
	gammaminus = Gamma(N-fEta);
	funeta= 1;//gammaminus/gammaplus*exp(-2.*EulerGamma*fEta);
	d0 = funeta;
	fSoft0 = funeta;
	double Ls = log(M*M/fmus/fmus);
	dcomplex d1, d2, d3, d4, d5, d6;
	dcomplex psi0m, psi0p, psi1m, psi1p, psi2m, psi2p, psi3m, psi3p, psi4m, psi4p, psi5m, psi5p;
	psi0m = psi0(N-fEta);
	psi0p = psi0(N+fEta);
	psi1m = psi_n(1,N-fEta);
	psi1p = psi_n(1,N+fEta);
	d1 = -(2*EulerGamma + psi0m + psi0p)*funeta;
	d2 = (pow(psi0m,2) + 4*EulerGamma*psi0p + pow(psi0p,2) +
			2*psi0m*(2*EulerGamma + psi0p) + psi1m - psi1p +
			4*pow(EulerGamma, 2))*funeta;
	fSoft1 = CA*(2.*d2+4.*Ls*d1+(M_PI*M_PI/3+2*Ls*Ls)*d0)*fasmus/4/M_PI;
	value =fSoft1;
	return value;
  }
  /**
   * One Loop Soft-Function Contribution
   */
  dcomplex SCETResummation::StildeOneLoop_LargeN(dcomplex N)
  {
	using namespace QCD;
	dcomplex Nbar = N*exp(EulerGamma);
	dcomplex Ls = log(M*M/fmus/fmus/Nbar/Nbar);
	dcomplex value;
	fEta = 0;//fEta1;
	fSoft1 = CA*(2.*Ls*Ls+M_PI*M_PI/3.)*fasmus/4/M_PI;
	value =fSoft1;
	value = value * pow(Nbar,-2*fEta);
	return value;
  }

  
  /**
    * Soft functions
    */
  dcomplex SCETResummation::SoftFunction(dcomplex N)
  {
    using namespace QCD;
    dcomplex gammaplus;
    dcomplex gammaminus;
    dcomplex funeta;
    dcomplex d0;
    dcomplex value;
    gammaplus = Gamma(N+fEta);
    gammaminus = Gamma(N-fEta);
    funeta= gammaminus/gammaplus*exp(-2.*EulerGamma*fEta);
    d0 = funeta;
    fSoft0 = funeta;
    double Ls = log(M*M/fmus/fmus);
    dcomplex d1, d2, d3, d4, d5, d6;
    dcomplex psi0m, psi0p, psi1m, psi1p, psi2m, psi2p, psi3m, psi3p, psi4m, psi4p, psi5m, psi5p;
    psi0m = psi0(N-fEta);
    psi0p = psi0(N+fEta);
    psi1m = psi_n(1,N-fEta);
    psi1p = psi_n(1,N+fEta);
    psi2m = psi_n(2,N-fEta);
    psi2p = psi_n(2,N+fEta);
    psi3m = psi_n(3,N-fEta);
    psi3p = psi_n(3,N+fEta);
    psi4m = psi_n(4,N-fEta);
    psi4p = psi_n(4,N+fEta);
    psi5m = psi_n(5,N-fEta);
    psi5p = psi_n(5,N+fEta);
    value = fH0*fULL*fSoft0;
    if(LL_order>0){
      d1 = -(2*EulerGamma + psi0m + psi0p)*funeta;
      d2 = (pow(psi0m,2) + 4*EulerGamma*psi0p + pow(psi0p,2) +
            2*psi0m*(2*EulerGamma + psi0p) + psi1m - psi1p +
            4*pow(EulerGamma, 2))*funeta;
      fSoft1 = CA*(2.*d2+4.*Ls*d1+(M_PI*M_PI/3+2*Ls*Ls)*d0)*fasmus/4/M_PI;
      value = (fH0)*(fUNLL)*(fSoft0); // NLL
      if(LogarithmicCounting==primed) value = (fH0+fH1)*(fUNLL)*(fSoft0+fSoft1); // NLL'
    }
    if(LL_order>1){
      d3=-(pow(psi0m,3) + 6*EulerGamma*pow(psi0p,2) + pow(psi0p,3) +
           3*pow(psi0m,2)* (2*EulerGamma + psi0p) + 6*EulerGamma*psi1m -
           6*EulerGamma*psi1p + psi2m + psi2p +
           3*psi0p*(psi1m - psi1p + 4*pow(EulerGamma, 2)) +
           3*psi0m*(4*EulerGamma*psi0p + pow(psi0p,2) + psi1m - psi1p +
                    4*pow(EulerGamma, 2)) + 8*pow(EulerGamma, 3))*funeta;
      d4=(pow(psi0m,4) + 8*EulerGamma*pow(psi0p,3)+ pow(psi0p,4) +
          4*pow(psi0m,3)*(2*EulerGamma + psi0p) + 3*pow(psi1m,2) - 6*psi1m*psi1p +
          3*pow(psi1p,2) + 8*EulerGamma*psi2m + 8*EulerGamma*psi2p + psi3m -
          psi3p + 24*psi1m*pow(EulerGamma, 2) - 24*psi1p*pow(EulerGamma, 2) +
          6*pow(psi0p,2)* (psi1m - psi1p + 4*pow(EulerGamma, 2)) +
          6*pow(psi0m,2)* (4*EulerGamma*psi0p + pow(psi0p,2) + psi1m - psi1p +
                           4*pow(EulerGamma, 2)) +
          4*psi0p*(6*EulerGamma*psi1m - 6*EulerGamma*psi1p + psi2m + psi2p +
                   8*pow(EulerGamma, 3)) +
          4*psi0m*(6*EulerGamma*pow(psi0p,2) + pow(psi0p,3) + 6*EulerGamma*psi1m -
                   6*EulerGamma*psi1p + psi2m + psi2p +
                   3*psi0p *(psi1m - psi1p + 4*pow(EulerGamma, 2)) +
                   8*pow(EulerGamma, 3)) + 16*pow(EulerGamma, 4))*funeta;
      fSoft2 = ((CA*CA*(2428./81-808.*Ls/27+134.*Ls*Ls/9-22./9*Ls*Ls*Ls+2.*Ls*Ls*Ls*Ls
                       +67./54*M_PI*M_PI-5./18*pow(M_PI,4)-22./9*ZETA3+28.*Ls*ZETA3)
                +CA*T_R*nf*(-656./81+224.*Ls/27-40.*Ls*Ls/9+8.*Ls*Ls*Ls/9-10.*M_PI*M_PI/27+8.*ZETA3/9))*d0+
               (CA*CA*(-808./27+268./9*Ls-22.*Ls*Ls/3+8.*Ls*Ls*Ls+28.*ZETA3)+
                CA*T_R*nf*(224./27-80./9*Ls+8./3*Ls*Ls))*d1+
               (CA*CA*(134./9-22.*Ls/3+12.*Ls*Ls)+CA*T_R*nf*(-40./9+8.*Ls/3))*d2+
               (CA*CA*(-22./9+8*Ls)+CA*T_R*nf*8./9)*d3+
               CA*CA*2*d4)*fasmus*fasmus/4/4/M_PI/M_PI;
      value = (fH0+fH1)*(fUNLL+fUNNLL)*(fSoft0+fSoft1); // NNLL
      if(LogarithmicCounting==primed)   value = (fH0+fH1+fH2)*(fUNLL+fUNNLL)*(fSoft0+fSoft1+fSoft2); // NNLL'
    }
	if (LL_order>2) {
	  d5 = -(((32*pow(EulerGamma,5) + pow(psi0m,5) + 10*EulerGamma*pow(psi0p,4) +
			   pow(psi0p,5) + 5*pow(psi0m,4)*(2*EulerGamma + psi0p) +
			   80*pow(EulerGamma,3)*psi1m + 30*EulerGamma*pow(psi1m,2) +
			   10*pow(psi0p,3)*(4*pow(EulerGamma,2) + psi1m - psi1p) +
			   10*pow(psi0m,3)*(4*pow(EulerGamma,2) + 4*EulerGamma*psi0p +
								  pow(psi0p,2) + psi1m - psi1p) - 80*pow(EulerGamma,3)*psi1p -
			   60*EulerGamma*psi1m*psi1p + 30*EulerGamma*pow(psi1p,2) +
			   40*pow(EulerGamma,2)*psi2m + 10*psi1m*psi2m - 10*psi1p*psi2m +
			   40*pow(EulerGamma,2)*psi2p + 10*psi1m*psi2p - 10*psi1p*psi2p +
			   10*pow(psi0p,2)*(8*pow(EulerGamma,3) + 6*EulerGamma*psi1m -
								  6*EulerGamma*psi1p + psi2m + psi2p) +
			   10*pow(psi0m,2)*(8*pow(EulerGamma,3) + 6*EulerGamma*pow(psi0p,2) +
								  pow(psi0p,3) + 6*EulerGamma*psi1m +
								  3*psi0p*(4*pow(EulerGamma,2) + psi1m - psi1p) - 6*EulerGamma*psi1p +
								  psi2m + psi2p) + 10*EulerGamma*psi3m +
			   5*psi0p*(16*pow(EulerGamma,4) + 3*pow(psi1m,2) +
						6*psi1m*(4*pow(EulerGamma,2) - psi1p) - 24*pow(EulerGamma,2)*psi1p +
						3*pow(psi1p,2) + 8*EulerGamma*psi2m + 8*EulerGamma*psi2p + psi3m - psi3p)\
			   + 5*psi0m*(16*pow(EulerGamma,4) + 8*EulerGamma*pow(psi0p,3) +
						  pow(psi0p,4) + 3*pow(psi1m,2) +
						  6*psi1m*(4*pow(EulerGamma,2) - psi1p) +
						  6*pow(psi0p,2)*(4*pow(EulerGamma,2) + psi1m - psi1p) -
						  24*pow(EulerGamma,2)*psi1p + 3*pow(psi1p,2) + 8*EulerGamma*psi2m +
						  8*EulerGamma*psi2p + 4*psi0p*
						  (8*pow(EulerGamma,3) + 6*EulerGamma*psi1m - 6*EulerGamma*psi1p + psi2m +
						   psi2p) + psi3m - psi3p) - 10*EulerGamma*psi3p + psi4m + psi4p)))*funeta;
			  d6 = ((64*pow(EulerGamma,6) + pow(psi0m,6) + 12*EulerGamma*pow(psi0p,5) +
					 pow(psi0p,6) + 6*pow(psi0m,5)*(2*EulerGamma + psi0p) +
					 240*pow(EulerGamma,4)*psi1m + 180*pow(EulerGamma,2)*pow(psi1m,2) +
					 15*pow(psi1m,3) + 15*pow(psi0p,4)*(4*pow(EulerGamma,2) + psi1m - psi1p) +
					 15*pow(psi0m,4)*(4*pow(EulerGamma,2) + 4*EulerGamma*psi0p + pow(psi0p,2) +
          psi1m - psi1p) - 240*pow(EulerGamma,4)*psi1p -
					 360*pow(EulerGamma,2)*psi1m*psi1p - 45*pow(psi1m,2)*psi1p +
					 180*pow(EulerGamma,2)*pow(psi1p,2) + 45*psi1m*pow(psi1p,2) -
					 15*pow(psi1p,3) + 160*pow(EulerGamma,3)*psi2m + 120*EulerGamma*psi1m*psi2m -
					 120*EulerGamma*psi1p*psi2m + 10*pow(psi2m,2) + 160*pow(EulerGamma,3)*psi2p +
					 120*EulerGamma*psi1m*psi2p - 120*EulerGamma*psi1p*psi2p + 20*psi2m*psi2p +
					 10*pow(psi2p,2) + 20*pow(psi0p,3)*
					 (8*pow(EulerGamma,3) + 6*EulerGamma*psi1m - 6*EulerGamma*psi1p + psi2m + psi2p)
					 + 20*pow(psi0m,3)*(8*pow(EulerGamma,3) + 6*EulerGamma*pow(psi0p,2) +
										  pow(psi0p,3) + 6*EulerGamma*psi1m +
										  3*psi0p*(4*pow(EulerGamma,2) + psi1m - psi1p) - 6*EulerGamma*psi1p + psi2m +
										  psi2p) + 60*pow(EulerGamma,2)*psi3m + 15*psi1m*psi3m - 15*psi1p*psi3m +
					 15*pow(psi0p,2)*(16*pow(EulerGamma,4) + 3*pow(psi1m,2) +
          6*psi1m*(4*pow(EulerGamma,2) - psi1p) - 24*pow(EulerGamma,2)*psi1p +
          3*pow(psi1p,2) + 8*EulerGamma*psi2m + 8*EulerGamma*psi2p + psi3m - psi3p) +
					 15*pow(psi0m,2)*(16*pow(EulerGamma,4) + 8*EulerGamma*pow(psi0p,3) +
          pow(psi0p,4) + 3*pow(psi1m,2) + 6*psi1m*(4*pow(EulerGamma,2) - psi1p) +
          6*pow(psi0p,2)*(4*pow(EulerGamma,2) + psi1m - psi1p) -
          24*pow(EulerGamma,2)*psi1p + 3*pow(psi1p,2) + 8*EulerGamma*psi2m +
          8*EulerGamma*psi2p + 4*psi0p*
										(8*pow(EulerGamma,3) + 6*EulerGamma*psi1m - 6*EulerGamma*psi1p + psi2m +
										 psi2p) + psi3m - psi3p) - 60*pow(EulerGamma,2)*psi3p - 15*psi1m*psi3p +
					 15*psi1p*psi3p + 12*EulerGamma*psi4m + 12*EulerGamma*psi4p +
					 6*psi0p*(32*pow(EulerGamma,5) + 30*EulerGamma*pow(psi1m,2) +
							  30*EulerGamma*pow(psi1p,2) + 40*pow(EulerGamma,2)*psi2m +
							  40*pow(EulerGamma,2)*psi2p -
							  10*psi1p*(8*pow(EulerGamma,3) + psi2m + psi2p) +
							  10*psi1m*(8*pow(EulerGamma,3) - 6*EulerGamma*psi1p + psi2m + psi2p) +
							  10*EulerGamma*psi3m - 10*EulerGamma*psi3p + psi4m + psi4p) +
					 6*psi0m*(32*pow(EulerGamma,5) + 10*EulerGamma*pow(psi0p,4) + pow(psi0p,5) +
							  30*EulerGamma*pow(psi1m,2) +
							  10*pow(psi0p,3)*(4*pow(EulerGamma,2) + psi1m - psi1p) -
							  80*pow(EulerGamma,3)*psi1p + 30*EulerGamma*pow(psi1p,2) +
							  40*pow(EulerGamma,2)*psi2m - 10*psi1p*psi2m + 40*pow(EulerGamma,2)*psi2p -
							  10*psi1p*psi2p + 10*psi1m*
							  (8*pow(EulerGamma,3) - 6*EulerGamma*psi1p + psi2m + psi2p) +
							  10*pow(psi0p,2)*(8*pow(EulerGamma,3) + 6*EulerGamma*psi1m -
             6*EulerGamma*psi1p + psi2m + psi2p) + 10*EulerGamma*psi3m +
							  5*psi0p*(16*pow(EulerGamma,4) + 3*pow(psi1m,2) +
									   6*psi1m*(4*pow(EulerGamma,2) - psi1p) - 24*pow(EulerGamma,2)*psi1p +
									   3*pow(psi1p,2) + 8*EulerGamma*psi2m + 8*EulerGamma*psi2p + psi3m - psi3p)\
							  - 10*EulerGamma*psi3p + psi4m + psi4p) + psi5m - psi5p))*funeta;
	  fSoft3 = (pow(CA,3)*((4*d6)/3. + (d5*(-2245320 + 3674160*Ls))/459270. +
							(d4*(-11226600*Ls + 9185400*pow(Ls,2) - 17010*(-925 + 108*ZETA2)))/
       459270. + (d3*(-22453200*pow(Ls,2) + 12247200*pow(Ls,3) -
					  68040*Ls*(-925 + 108*ZETA2) + 22680*(-2102 + 99*ZETA2 + 1134*ZETA3)
					  ))/459270. + (d2*(-22453200*pow(Ls,3) + 9185400*pow(Ls,4) -
										102060*pow(Ls,2)*(-925 + 108*ZETA2) +
										68040*Ls*(-2102 + 99*ZETA2 + 1134*ZETA3) -
										1134*(-101795 + 6030*ZETA2 + 4212*pow(ZETA2,2) + 37620*ZETA3)))/
       459270. + (d1*(-11226600*pow(Ls,4) + 3674160*pow(Ls,5) -
					  68040*pow(Ls,3)*(-925 + 108*ZETA2) +
					  68040*pow(Ls,2)*(-2102 + 99*ZETA2 + 1134*ZETA3) -
					  2268*Ls*(-101795 + 6030*ZETA2 + 4212*pow(ZETA2,2) +
							  37620*ZETA3) + 126*
					  (-1485145 + 69480*ZETA2 + 64152*pow(ZETA2,2) + 1664280*ZETA3 -
					   9720*ZETA2*ZETA3 - 699840*ZETA5)))/459270. +
							(d0*(-2245320*pow(Ls,5) + 612360*pow(Ls,6) -
								 57983688*pow(ZETA2,2) + 22661208*pow(ZETA2,3) -
								 17010*pow(Ls,4)*(-925 + 108*ZETA2) +
								 22680*pow(Ls,3)*(-2102 + 99*ZETA2 + 1134*ZETA3) -
								 630*ZETA2*(-23333 + 21384*ZETA3) -
								 1134*pow(Ls,2)*(-101795 + 6030*ZETA2 + 4212*pow(ZETA2,2) +
              37620*ZETA3) + 35*
								 (5211949 - 4700808*ZETA3 + 1562976*pow(ZETA3,2) -
								  1411344*ZETA5) +
								 126*Ls*(-1485145 + 69480*ZETA2 + 64152*pow(ZETA2,2) +
										1664280*ZETA3 - 9720*ZETA2*ZETA3 - 699840*ZETA5)))/459270.) +
   pow(CA,2)*nf*((8*d5)/9. + (d4*(-2789640 + 2041200*Ls))/459270. +
				   (d3*(-11158560*Ls + 4082400*pow(Ls,2) - 22680*(-457 + 18*ZETA2)))/
       459270. + (d2*(-16737840*pow(Ls,2) + 4082400*pow(Ls,3) -
					  68040*Ls*(-457 + 18*ZETA2) + 34020*(-793 + 30*ZETA2 + 12*ZETA3)))/
       459270. + (d1*(-11158560*pow(Ls,3) + 2041200*pow(Ls,4) -
					  68040*pow(Ls,2)*(-457 + 18*ZETA2) +
					  252*(156565 - 3420*ZETA2 + 5832*pow(ZETA2,2) - 61020*ZETA3) +
					  68040*Ls*(-793 + 30*ZETA2 + 12*ZETA3)))/459270. +
				   (d0*(-2789640*pow(Ls,4) + 408240*pow(Ls,5) - 2057580*ZETA2 +
						5293512*pow(ZETA2,2) - 22680*pow(Ls,3)*(-457 + 18*ZETA2) +
						252*Ls*(156565 - 3420*ZETA2 + 5832*pow(ZETA2,2) - 61020*ZETA3) +
						34020*pow(Ls,2)*(-793 + 30*ZETA2 + 12*ZETA3) +
						70*(-412765 + 98496*ZETA3 - 34992*ZETA5)))/459270.) +
   CA*(pow(nf,2)*((4*d4)/27. + (d3*(-453600 + 272160*Ls))/459270. +
					(d2*(1134000 - 1360800*Ls + 408240*pow(Ls,2)))/459270. +
					(d1*(2268000*Ls - 1360800*pow(Ls,2) + 272160*pow(Ls,3) -
						 20160*(58 + 81*ZETA3)))/459270. +
					(d0*(-17920 + 1134000*pow(Ls,2) - 453600*pow(Ls,3) +
						 68040*pow(Ls,4) - 45360*ZETA2 + 353808*pow(ZETA2,2) +
						 1663200*ZETA3 - 20160*Ls*(58 + 81*ZETA3)))/459270.) +
	   CF*nf*((4*d3)/3. + (d2*(1837080*Ls + 153090*(-55 + 48*ZETA3)))/
			  459270. + (d1*(1837080*pow(Ls,2) + 306180*Ls*(-55 + 48*ZETA3) -
							 3402*(-8555 + 864*pow(ZETA2,2) + 4560*ZETA3)))/459270. +
			  (d0*(-40377015 + 612360*pow(Ls,3) + 3102624*pow(ZETA2,2) +
				   16102800*ZETA3 + 153090*pow(Ls,2)*(-55 + 48*ZETA3) +
				   51030*ZETA2*(-55 + 48*ZETA3) -
				   3402*Ls*(-8555 + 864*pow(ZETA2,2) + 4560*ZETA3) +
				   11430720*ZETA5))/459270.)) )*fasmus*fasmus*fasmus/4/4/4/M_PI/M_PI/M_PI;
	  value =  (fH0+fH1+fH2)*(fUNLL+fUNNLL+fUNNNLL)*(fSoft0+fSoft1+fSoft2); // NNNLL
	  if(LogarithmicCounting==primed) value =  (fH0+fH1+fH2+fH3)*(fUNLL+fUNNLL+fUNNNLL)*(fSoft0+fSoft1+fSoft2+fSoft3); // NNNLL'
	}
	return value;
  }

  dcomplex SCETResummation::SoftFunction_LargeN(dcomplex N) {
    using namespace QCD;
    dcomplex Nbar = N*exp(EulerGamma);
    dcomplex Ls = log(M*M/fmus/fmus/Nbar/Nbar);
    dcomplex value;
    fSoft0 = 1.;
    value = fH0*fULL*fSoft0;
    if(LL_order>0){
      fSoft1 = CA*(2.*Ls*Ls+M_PI*M_PI/3.)*fasmus/4/M_PI;
      value = (fH0)*(fUNLL)*(fSoft0); // NLL
      if(LogarithmicCounting==primed) value = (fH0+fH1)*(fUNLL)*(fSoft0+fSoft1);
    }
    if(LL_order>1){
      fSoft2= fasmus*fasmus/4/4/M_PI/M_PI*(
					   CA*CA*(2*Ls*Ls*Ls*Ls-22./9*Ls*Ls*Ls+134./9*Ls*Ls+(-808./27+28.*ZETA3)*Ls+2428./81+67.*M_PI*M_PI/54-5.*M_PI*M_PI*M_PI*M_PI/18-22./9*ZETA3)
					   +CA*T_R*nf*(8./9*Ls*Ls*Ls-40./9*Ls*Ls+224./27*Ls-656./81-10.*M_PI*M_PI/27+8./9*ZETA3));
      value =  (fH0+fH1)*(fUNLL+fUNNLL)*(fSoft0+fSoft1); // NNLL
      if(LogarithmicCounting==primed) value = (fH0+fH1+fH2)*(fUNLL+fUNNLL)*(fSoft0+fSoft1+fSoft2);
    }
    if(LL_order>2){
      fSoft3 = fasmus*fasmus*fasmus/4/4/4/M_PI/M_PI/M_PI*(
							  pow(CA,2)*nf*((-164*pow(Ls,4))/27. + (8*pow(Ls,5))/9. -
									    (4*pow(Ls,3)*(-457 + 18*ZETA2))/81. +
									(2*Ls*(156565 - 3420*ZETA2 + 5832*pow(ZETA2,2) - 61020*ZETA3))/
									3645. + (2*pow(Ls,2)*(-793 + 30*ZETA2 + 12*ZETA3))/27. +
									(-2057580*ZETA2 + 5293512*pow(ZETA2,2) +
									 70*(-412765 + 98496*ZETA3 - 34992*ZETA5))/459270.) +
							  pow(CA,3)*((-44*pow(Ls,5))/9. + (4*pow(Ls,6))/3. -
								     (pow(Ls,4)*(-925 + 108*ZETA2))/27. +
								     (4*pow(Ls,3)*(-2102 + 99*ZETA2 + 1134*ZETA3))/81. -
								     (pow(Ls,2)*(-101795 + 6030*ZETA2 + 4212*pow(ZETA2,2) +
										 37620*ZETA3))/405. +
								     (-57983688*pow(ZETA2,2) + 22661208*pow(ZETA2,3) -
								      630*ZETA2*(-23333 + 21384*ZETA3) +
								      35*(5211949 - 4700808*ZETA3 + 1562976*pow(ZETA3,2) -
									  1411344*ZETA5))/459270. +
								     (Ls*(-1485145 + 69480*ZETA2 + 64152*pow(ZETA2,2) + 1664280*ZETA3 -
									  9720*ZETA2*ZETA3 - 699840*ZETA5))/3645.) +
							  CA*(pow(nf,2)*((200*pow(Ls,2))/81. - (80*pow(Ls,3))/81. +
									 (4*pow(Ls,4))/27. - (32*Ls*(58 + 81*ZETA3))/729. +
									 (-17920 - 45360*ZETA2 + 353808*pow(ZETA2,2) + 1663200*ZETA3)/
									 459270.) + CF*nf*((4*pow(Ls,3))/3. +
											   (pow(Ls,2)*(-55 + 48*ZETA3))/3. -
											   (Ls*(-8555 + 864*pow(ZETA2,2) + 4560*ZETA3))/135. + 
											   (-40377015 + 3102624*pow(ZETA2,2) + 16102800*ZETA3 + 
											    51030*ZETA2*(-55 + 48*ZETA3) + 11430720*ZETA5)/459270.)));
      value =  (fH0+fH1+fH2)*(fUNLL+fUNNLL+fUNNNLL)*(fSoft0+fSoft1+fSoft2); // NNNLL
      if(LogarithmicCounting==primed) value =  (fH0+fH1+fH2+fH3)*(fUNLL+fUNNLL+fUNNNLL)*(fSoft0+fSoft1+fSoft2+fSoft3); // NNNLL'
    }
    value = value * pow(Nbar,-2*fEta);
    return value;
  }

  
  /**
    * Soft Function times Luminosity in N space
    */
  dcomplex SCETResummation::lum_times_SoftFunction(dcomplex N)
  {
	dcomplex result;
	if (APimpr>2) {
	  cout << "\033[1;35m********************WARNING********************" << endl << "AP improvement not supported! Using AP0 instead" << endl << "***********************************************\033[0m" << endl;
	  APimpr = AP0;
	}

	if(prescr == BP2)
	{
	  if (APimpr == AP2 )result = (2*SoftFunction(N)-3*SoftFunction(N+1)+2*SoftFunction(N+2));
	  else if (APimpr == AP1 )result = SoftFunction(N+1);
	  else result = SoftFunction(N);
	}
	else if(prescr == MP00) result = SoftFunction_LargeN(N-0.5);
	else if(prescr == MP2)
	{
	  if (APimpr == AP2 ) result = (2*SoftFunction_LargeN(exp(psi0(N)))-3*SoftFunction_LargeN(exp(psi0(N+1)))+2*SoftFunction_LargeN(exp(psi0(N+2))));
	  else if (APimpr == AP1 ) result = SoftFunction_LargeN(exp(psi0(N+1)));
	  else result = SoftFunction_LargeN(exp(psi0(N)));
	}
	  else{
		cout << "\033[1;31m***************ERROR****************" << endl <<
		"Prescription not implemented! Abort" << endl << "************************************\033[0m" << endl;
		abort();
	}
	//
	if(APimpr == 0) result -= Subtraction(N);
	if(APimpr == 1) result -= Subtraction(N+1);
	if(APimpr == 2) result -= 2*Subtraction(N) - 3*Subtraction(N+1) + 2*Subtraction(N+2);
	//
	return result*LuminosityN(N);
  }
  
  /**
    * Integrand
    */
  double SCETMinimal_integrand(double t, void *p)
  {
    SCETResummation *Mres = (SCETResummation *) p;
    double x = Mres->tau;
    dcomplex N = Mres->MP_reN0 + (Mres->MP_slope-I)*log(t);
    dcomplex res = (I-Mres->MP_slope) * exp(-N*log(x)) * Mres->lum_times_SoftFunction(N)/t;
    return imag(res)/M_PI;
  }
  
  /**
    * Minimal Prescription Implementation
    */
  double SCETResummation::SCETMinimalPrescription()
  {
    double prec = 1e-4;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    return integration::gauss(SCETMinimal_integrand,0.,1.,prec,NULL,this);
    //return integration::gauss_inf(SCETMinimal_integrand_inf,0.,prec,NULL,&x);  // fails..
    gsl_set_error_handler (old_handler);
  }
  
  /**
   * Soft Function times Luminosity in N space
   */
  dcomplex SCETResummation::lum_times_Stilde(dcomplex N)
  {
	dcomplex result;
	if(prescr == BP2)
	{
	  if (APimpr == AP2 )result = (2*StildeOneLoop(N)-3*StildeOneLoop(N+1)+2*StildeOneLoop(N+2));
	  else if (APimpr == AP1 )result = StildeOneLoop(N+1);
	  else result = StildeOneLoop(N);
	}
	else if(prescr == MP00) result = StildeOneLoop_LargeN(N);
	else if(prescr == MP2)
	{
	  if (APimpr == AP2 ) result = (2*StildeOneLoop_LargeN(exp(psi0(N)))-3*StildeOneLoop_LargeN(exp(psi0(N+1)))+2*StildeOneLoop_LargeN(exp(psi0(N+2))));
	  else if (APimpr == AP1 ) result = StildeOneLoop_LargeN(exp(psi0(N+1)));
	  else result = StildeOneLoop_LargeN(exp(psi0(N)));
	}
	else{
	  cout << "\033[1;31m***************ERROR****************" << endl <<
	  "Prescription not implemented! Abort" << endl << "************************************\033[0m" << endl;
	  abort();
	}
	return result*LuminosityN(N);
	//return StildeOneLoop(N)*LuminosityN(N);
  }

  /**
   * Integrand
   */
  double StildeMinimal_integrand(double t, void *p)
  {
	SCETResummation *Mres = (SCETResummation *) p;
	double x = Mres->tau;
	dcomplex N = Mres->MP_reN0 + (Mres->MP_slope-I)*log(t);
	dcomplex res = (I-Mres->MP_slope) * exp(-N*log(x)) * Mres->lum_times_Stilde(N)/t;
	return imag(res)/M_PI;
  }
  
  /**
   * Minimal Prescription Implementation
   */
  double SCETResummation::StildeMinimalPrescription()
  {
	double prec = 1e-4;
	gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
	gsl_set_error_handler_off();
	return integration::gauss(StildeMinimal_integrand,0.,1.,prec,NULL,this);
	//return integration::gauss_inf(SCETMinimal_integrand_inf,0.,prec,NULL,&x);  // fails..
	gsl_set_error_handler (old_handler);
  }
  
  /**
    * SCET coefficients and constants initialisation
    */
  void SCETResummation::InitResummCoeffsSCET()
  {
    if(proc >= 1)
    {  // Drell-Yan
      if(!(last_nf==nf)) InitCoeffDY();
      InitConstDY();
    } else if(proc == ggH)
    {  // Higgs
      //if(!(last_nf==nf))
      InitCoeffHiggsSCET();
      //InitConstHiggsSCET();
    }
    return;
  }
 
  /**
    * Hard Function
    */
  void SCETResummation::H()
  {
    InitConstHiggsCommons();  // initialize Wilson coefficient, and equivalent for finite mt
    dcomplex fCt1, fCt2, fCt3;
    fCt1 = W1/2;
    fCt2 = W2/2 - W1*W1/8;
    fCt3 = W3/2 - W1*W2/4 + W1*W1*W1/16;
    if(fmut != M) {
      dcomplex lHt  = log(fmut*fmut/mtop/mtop);
      double   lHt0 = 2*log(M/mtop);
      fCt2 += (19./16. + 1./3.*nf) * (lHt-lHt0);
      fCt3 = -2892659./41472. + 1733./288.*lHt + 209./64.*lHt*lHt + 897943./9216.*ZETA3
	+ nf* ( 40291./20736. + 55./54.*lHt + 23./32.*lHt*lHt - 110779./13824.*ZETA3 )
	+ nf*nf* ( - 6865./31104. + 77./1728.*lHt - 1./18.*lHt*lHt );
    }
    fCt1 /= M_PI;
    fCt2 /= M_PI*M_PI;
    fCt3 /= M_PI*M_PI*M_PI;

    //cout << fCt1 << "   "<< fCt2 << "   "<< fCt3 << "   "<< endl;
    using namespace QCD;

    dcomplex CS1, CS2, CS3;
    dcomplex L = log(dcomplex(-M*M/fmuh/fmuh));
    //cout << L << endl;

    CS1 = dcomplex(CA*(-L*L+M_PI*M_PI/6.)/4./M_PI);
    CS2 = (CA*CA*(L*L*L*L/2.+11./9*L*L*L+(-67./9+M_PI*M_PI/6)*L*L+(80./27-11./9*M_PI*M_PI-2.*ZETA3)*L
		  +5105./162+67.*M_PI*M_PI/36+M_PI*M_PI*M_PI*M_PI/72-143./9*ZETA3)+CF*T_R*nf*(4.*L-67./3+16.*ZETA3)
	   +CA*T_R*nf*(-4./9*L*L*L+20./9*L*L+(104./27+4.*M_PI*M_PI/9)*L-1832./81-5*M_PI*M_PI/9-92./9*ZETA3))/4/4/M_PI/M_PI;
    CS3 = (CA*pow(nf,2)*(46.59358329522939 + (40*pow(L,3))/81. - (2*pow(L,4))/27. +
			 pow(L,2)*(1.4320987654320987 + (8*ZETA2)/9.) + (4*ZETA2)/9. +
			 (4*pow(ZETA2,2))/27. + L*(-19.282578875171467 - (80*ZETA2)/27. -
						   (128*ZETA3)/27.) + (4576*ZETA3)/243.) +
	   CF*pow(nf,2)*(55.32098765432099 + (4*pow(L,2))/3. - (20*ZETA2)/9. -
			 (16*pow(ZETA2,2))/45. - (112*ZETA3)/3. + L*(-17.333333333333332 + (32*ZETA3)/3.))
	   + pow(CF,2)*nf*(33.77777777777778 - 2*L + (296*ZETA3)/3. - 160*ZETA5) +
	   pow(CA,2)*nf*(-573.8465172991922 - (8*pow(L,4))/27. + (2*pow(L,5))/9. +
			 pow(L,3)*(-9.061728395061728 - (2*ZETA2)/3.) - (14189*ZETA2)/729. +
			 (73*pow(ZETA2,2))/45. - (460*ZETA3)/81. - (82*ZETA2*ZETA3)/9. +
			 L*(182.49108367626886 + (3820*ZETA2)/81. - (48*pow(ZETA2,2))/5. +
			    (28*ZETA3)/9.) + pow(L,2)*
			 (13.962962962962964 - (103*ZETA2)/9. + (118*ZETA3)/9.) + (428*ZETA5)/9.) +
	   CA*CF*nf*(-351.048353909465 - (8*pow(L,3))/3. - (68*ZETA2)/9. -
		     (64*pow(ZETA2,2))/45. + L*(70.98148148148148 + 6*ZETA2 +
						(16*pow(ZETA2,2))/5. - (376*ZETA3)/9.) + pow(L,2)*(13 - 16*ZETA3) +
		     (14564*ZETA3)/81. + (64*ZETA2*ZETA3)/3. + (608*ZETA5)/9.) +
	   pow(CA,3)*(1129.3733043743332 - (11*pow(L,5))/9. - pow(L,6)/6. +
		      pow(L,4)*(5.203703703703703 - (3*ZETA2)/2.) + (105617*ZETA2)/729. -
		      (1939*pow(ZETA2,2))/270. - (24389*pow(ZETA2,3))/1890. - (152716*ZETA3)/243. -
		      (605*ZETA2*ZETA3)/9. - (104*pow(ZETA3,2))/9. +
		      pow(L,3)*(19.012345679012345 + (11*ZETA2)/3. + 2*ZETA3) +
		      pow(L,2)*(-83.20987654320987 + (685*ZETA2)/18. - (73*pow(ZETA2,2))/10. +
				(143*ZETA3)/9.) + (2222*ZETA5)/9. +
		      L*(-256.4986282578875 - (13420*ZETA2)/81. + (176*pow(ZETA2,2))/5. +
			 (2048*ZETA3)/27. + (34*ZETA2*ZETA3)/3. + 16*ZETA5)))/4./4./4./M_PI/M_PI/M_PI;
    
    //cout << CS1 << "   "<< CS2 << "   "<< CS3 << "   "<< endl;
    // compute Ct*CS
    CS3 = fasmuh*fasmuh*fasmuh*CS3 + fasmuh*fasmuh*fasmut*fCt1*CS2 + fasmuh*fasmut*fasmut*fCt2*CS1 + fasmut*fasmut*fasmut*fCt3;
    CS2 = fasmuh*fasmuh*CS2 + fasmuh*fasmut*fCt1*CS1 + fasmut*fasmut*fCt2;
    CS1 = fasmuh*CS1 + fasmut*fCt1;
    
    //cout << CS1 << "   "<< CS2 << "   "<< CS3 << "   "<< endl;
    
    fH0 = 1;
    fH1 = 2*real(CS1);
    fH2 = 2*real(CS2) + norm(CS1); // norm = abs^2
    fH3 = 2*real(CS3) + 2*real(CS1*CS2);
    
    //cout << fH1 << "   "<< fH2 << "   "<< fH3 << "   "<< endl;
    
  }

  /**
    * Eta
    */
  void SCETResummation::Eta()
  {
    fEta1 = 2.*fGammaA0/2./fbbeta0*(log(fasmuf/fasmus));
    fEta2 = 2.*fGammaA0/2./fbbeta0*((fGammaA1/fGammaA0-fbbeta1/fbbeta0)*(fasmuf-fasmus)/4./M_PI);
    fEta3 = 2.*fGammaA0/2./fbbeta0*(fGammaA2/fGammaA0
				    -fbbeta2/fbbeta0-fbbeta1/fbbeta0*(fGammaA1/fGammaA0-fbbeta1/fbbeta0))*(fasmuf*fasmuf-fasmus*fasmus)/32./M_PI/M_PI;
    fEta4 = (fGammaA3 - fGammaA2*fbbeta1/fbbeta0
	     + (fbbeta1*fbbeta1/fbbeta0/fbbeta0 - fbbeta2/fbbeta0) * fGammaA1
	     + (-pow(fbbeta1,3)/pow(fbbeta0,3) + 2*fbbeta2*fbbeta1/pow(fbbeta0,2)
		- fbbeta3/fbbeta0) * fGammaA0) * (pow(fasmuf,3) -pow(fasmus,3))/3/pow(4.*M_PI,3);
    //
    fEta = fEta1;
    if(LL_order > 0){
      fEta += fEta2;
    }
    if(LL_order > 1 ){
      fEta += fEta3;
    }
    if(LL_order > 2 ){
      fEta += fEta4;
    }
    //cout << fEta1 << "   "<< fEta2 << "   "<< fEta3 << "   "<< endl;
  }
  
  /**
    * Evolution
    */
  void SCETResummation::InitAlphas()
  {
    //this definition has to be completed at 4 loops
    if(fPi==false) fasmuh = alphas(real(fmuh));
    else
      {
		double asmuh_im = alphas(imag(fmuh));
		double a = fbbeta0*asmuh_im/4.;
		dcomplex X = 1. + I*a;
		fasmuh = alphas(imag(fmuh))/X;
		if(LL_order > 0)
		  fasmuh += -b1*asmuh_im*asmuh_im/X/X*log(X);
		if (LL_order > 1) {
		  fasmuh += pow(asmuh_im,3)/pow(X,3)*(b1*b1*(pow(log(X),2)-log(X)-1+X)+b2*(1-X));
		}
		if (LL_order > 2){
		  fasmuh += pow(asmuh_im,4)*(-((-1 + X)*(pow(b1,3)*(-1 + X) - 2*b1*b2*X + b3*(1 + X)) +
									   b1*(6*b2 + 4*pow(b1,2)*(-1 + X) - 4*b2*X)*log(X) -
									   5*pow(b1,3)*pow(log(X),2) + 2*pow(b1,3)*pow(log(X),3))/(2.*pow(X,4)));
		}
	/*
	  double imagas = alphas(imag(muh));
	  double a = fbbeta0*imagas/4.;
	  fasmuh = imagas/(1. - I*a);
	  //check if the following definition is ok
	  if(LL_order>=1)
	  {
	  fasmuh = imagas/(1. - I*a+fbbeta1/fbbeta0*imagas/4/M_PI*log(dcomplex(1.-I*a)));
	  }
	*/
      }

    fasmut = fasmuh;

    if(mtEFT) fasmut = alphas(real(fmut));


    fasmuf = alphas(muF);
    fasmus = alphas(fmus);
    //
    //cout << fasmuf << endl;
    //cout << fasmus << endl;
    //cout << fasmut << "  " << fmut<< endl;
    //cout << fasmuh << endl;
    //
    return;
  }	

  void SCETResummation::U()
  {
	faGA0 = 0;
	faGA1 = fGammaA0/2./fbbeta0*(log(dcomplex(fasmus/fasmuh)));
	faGA2 = fGammaA0/2./fbbeta0*((fGammaA1/fGammaA0-fbbeta1/fbbeta0)*(fasmus-fasmuh)/4./M_PI);
	faGA3 = fGammaA0/2./fbbeta0*(fGammaA2/fGammaA0
				     -fbbeta2/fbbeta0-fbbeta1/fbbeta0*(fGammaA1/fGammaA0-fbbeta1/fbbeta0))*(fasmus*fasmus-fasmuh*fasmuh)/32./M_PI/M_PI;
	
	faGS0 = 0;
	faGS1 = fGammaS0/2./fbbeta0*(log(dcomplex(fasmus/fasmuh)));
	faGS2 = 1./2./fbbeta0*((fGammaS1-fGammaS0*fbbeta1/fbbeta0)*(fasmus-fasmuh)/4./M_PI);
	faGS3 = 1./2./fbbeta0*(fGammaS2
			       -fGammaS0*fbbeta2/fbbeta0-fbbeta1/fbbeta0*(fGammaS1-fGammaS0*fbbeta1/fbbeta0))*(fasmus*fasmus-fasmuh*fasmuh)/32./M_PI/M_PI;
	
	faGB0 = 0;
	faGB1 = fGammaB0/2./fbbeta0*(log(dcomplex(fasmuf/fasmus)))+1./2*log(fasmus/fasmuf);
	faGB2 = fGammaB0/2./fbbeta0*((fGammaB1/fGammaB0-fbbeta1/fbbeta0)*(fasmuf-fasmus)/4./M_PI);
	faGB3 = fGammaB0/2./fbbeta0*(fGammaB2/fGammaB0
				     -fbbeta2/fbbeta0-fbbeta1/fbbeta0*(fGammaB1/fGammaB0-fbbeta1/fbbeta0))*(fasmuf*fasmuf-fasmus*fasmus)/32./M_PI/M_PI;
	
	faGT0 = 0;
	faGT1 = fGammaT0/2./fbbeta0*(log(dcomplex(fasmus/fasmut)));
	faGT2 = 1./2./fbbeta0*((fGammaT1-fGammaT0*fbbeta1/fbbeta0)*(fasmus-fasmut)/4./M_PI);
	faGT3 = 1./2./fbbeta0*(fGammaT2
			       -fGammaT0*fbbeta2/fbbeta0-fbbeta1/fbbeta0*(fGammaT1-fGammaT0*fbbeta1/fbbeta0))*(fasmus*fasmus-fasmut*fasmut)/32./M_PI/M_PI;
	
	dcomplex r = dcomplex(fasmus/fasmuh);
	
	fS0=fGammaA0/4/fbbeta0/fbbeta0*(4.*M_PI/fasmuh*(1.-1./r-log(r)));
	fS1=fGammaA0/4/fbbeta0/fbbeta0*((fGammaA1/fGammaA0-fbbeta1/fbbeta0)*(1.-r+log(r))+fbbeta1/2./fbbeta0*log(r)*log(r));
	fS2=fGammaA0/4/fbbeta0/fbbeta0*fasmuh/4./M_PI*((fbbeta1*fGammaA1/fbbeta0/fGammaA0-fbbeta2/fbbeta0)*(1.-r+r*log(r))+(fbbeta1*fbbeta1/fbbeta0/fbbeta0-fbbeta2/fbbeta0)*(1.-r)*log(r)
						       -(fbbeta1*fbbeta1/fbbeta0/fbbeta0-fbbeta2/fbbeta0-fbbeta1*fGammaA1/fbbeta0/fGammaA0+fGammaA2/fGammaA0)*(1.-r)*(1.-r)/2.);
	fS3 = fGammaA0/4/fbbeta0/fbbeta0*(fasmuh/4./M_PI)*(fasmuh/4./M_PI)*((fbbeta1*fbbeta2/fbbeta0/fbbeta0-fbbeta1*fbbeta1*fbbeta1/2./fbbeta0/fbbeta0/fbbeta0-fbbeta3/2./fbbeta0
									     +fbbeta1/fbbeta0*(fGammaA2/fGammaA0-fbbeta2/fbbeta0+fbbeta1*fbbeta1/fbbeta0/fbbeta0-fbbeta1*fGammaA1/fbbeta0/fGammaA0)*r*r/2.)*log(r)
									    +(fGammaA3/fGammaA0-fbbeta3/fbbeta0+2.*fbbeta1*fbbeta2/fbbeta0/fbbeta0+fbbeta1*fbbeta1/fbbeta0/fbbeta0*(fGammaA1/fGammaA0-fbbeta1/fbbeta0)
									      -fbbeta2*fGammaA1/fbbeta0/fGammaA0-fbbeta1/fbbeta0*fGammaA2/fGammaA0)*(1.-r)*(1.-r)*(1.-r)/3.
									    +(3.*fbbeta3/4./fbbeta0-fGammaA3/2./fGammaA0+fbbeta1*fbbeta1*fbbeta1/fbbeta0/fbbeta0/fbbeta0-3./4.*fbbeta1*fbbeta1/fbbeta0/fbbeta0*fGammaA1/fGammaA0
									      +fbbeta2/fbbeta0*fGammaA1/fGammaA0+fbbeta1/4./fbbeta0*fGammaA2/fGammaA0-7./4.*fbbeta1*fbbeta2/fbbeta0/fbbeta0)*(1.-r)*(1.-r)
									    +(fbbeta1*fbbeta2/fbbeta0/fbbeta0-fbbeta3/fbbeta0-fbbeta1*fbbeta1*fGammaA1/fbbeta0/fbbeta0/fGammaA0+fbbeta1/fbbeta0*fGammaA2/fGammaA0)*(1.-r)/2.);
	
	
	//2*log(fasmus/fasmuF) is added to faGB1
	dcomplex Lh, U0, U1, U2, U3;
	Lh = log(dcomplex((-M*M-I*0.0001)/fmuh/fmuh));
	
	
	U0 = 4.*fS0-2.*faGS0-2.*faGT0+4.*faGB0-2.*faGA0*Lh;
	U1 = 4.*fS1-2.*faGS1-2.*faGT1+4.*faGB1-2.*faGA1*Lh;
	U2 = 4.*fS2-2.*faGS2-2.*faGT2+4.*faGB2-2.*faGA2*Lh;
	U3 = 4.*fS3-2.*faGS3-2.*faGT3+4.*faGB3-2.*faGA3*Lh;
	
	fULL = real(exp(real(U0)));
	fUNLL = real(exp(real(U0+U1)));
	fUNNLL = real(fUNLL*(real(U2)));
	fUNNNLL = real(fUNLL*(real(U2)*real(U2)/2.+real(U3)));

	//cout << fULL << "   "<< fUNLL << "   "<< fUNNLL << "   " << fUNNNLL<< endl;
	
  }
  
  /**
    * Coefficients
    */
  void SCETResummation::InitCoeffHiggsSCET()
  {
    using namespace QCD;
    InitCoeffDY();
    fGammaA0 = 4*M_PI*A1*CA/CF;
    fGammaA1 = 16*M_PI*M_PI*A2*CA/CF;
    fGammaA2 = 64*M_PI*M_PI*M_PI*A3*CA/CF;
    fGammaA3 = 256*M_PI*M_PI*M_PI*M_PI*A4*CA/CF;
    fbbeta0 = b0*4*M_PI;
    fbbeta1 = b1*b0*16*M_PI*M_PI;
    fbbeta2 = b2*b0*64*M_PI*M_PI*M_PI;
    fbbeta3 = b3*b0*256*M_PI*M_PI*M_PI*M_PI;
    fGammaS0=0;
    fGammaS1=CA*CA*(-160./27.+11.*M_PI*M_PI/9.+4.*1.2020569031595942855)+CA*T_R*nf*(-208./27.-4.*M_PI*M_PI/9.)-8.*CF*T_R*nf;
    fGammaS2=CA*CA*CA*(37045./729.+6109.*M_PI*M_PI/243-319.*pow(M_PI,4)/135+(244./3-40.*M_PI*M_PI/9)*1.2020569031595942855-32.*1.036927755143369926331)
      +CA*CA*T_R*nf*(-167800./729-2396.*M_PI*M_PI/243+164.*pow(M_PI,4)/135+1424./27*1.2020569031595942855)
      +CA*CF*T_R*nf*(1178./27-4.*M_PI*M_PI/3-16.*pow(M_PI,4)/45-608./9*1.2020569031595942855)+8.*CF*CF*T_R*nf
      +CA*T_R*T_R*nf*nf*(24520./729+80.*M_PI*M_PI/81-448./27*1.2020569031595942855)+176./9*CF*T_R*T_R*nf*nf;
    fGammaB0=11./3.*CA-4./3.*T_R*nf;
    fGammaB1=4.*CA*CA*(8./3.+3.*1.2020569031595942855)-16./3.*CA*T_R*nf-4.*CF*T_R*nf;
    fGammaB2=CA*CA*CA*(79./2+4.*M_PI*M_PI/9+11.*pow(M_PI,4)/54+(536./3-8.*M_PI*M_PI/3)*1.2020569031595942855-80.*1.036927755143369926331)
      +CA*CA*T_R*nf*(233./9+8.*M_PI*M_PI/9+2.*pow(M_PI,4)/27+160./3*1.2020569031595942855)
      -241./9*CA*CF*T_R*nf+2.*CF*CF*T_R*nf+58./9*CA*T_R*T_R*nf*nf+44./9*CF*T_R*T_R*nf*nf;
    fGammaT0=0;
    fGammaT1=-2.*fbbeta1;
    fGammaT2=-4.*fbbeta2/2.;
    //cout << fGammaA0 << "   "<< fGammaA1 << "   "<< fGammaA2 << "   "<<fGammaA3<< endl;
    //cout << fbbeta0 << "   "<< fbbeta1 << "   "<< fbbeta2 << "   "<<fbbeta3<< endl;
    //cout << fGammaS0 << "   "<< fGammaS1 << "   "<< fGammaS2 << "   "<< endl;
    //cout << fGammaB0 << "   "<< fGammaB1 << "   "<< fGammaB2 << "   "<< endl;
    //cout << fGammaT0 << "   "<< fGammaT1 << "   "<< fGammaT2 << "   "<< endl;
    return;
  }
  
  /**
    * Constants
    */
  /*
  void SCETResummation::InitConstHiggsSCET()
  {
    using namespace QCD;

    dcomplex L = log(dcomplex(-M*M/fmuh/fmuh));
	
    CS1 = dcomplex(CA*(-L*L+M_PI*M_PI/6.)/4./M_PI);
    CS2 = (CA*CA*(L*L*L*L/2.+11./9*L*L*L+(-67./9+M_PI*M_PI/6)*L*L+(80./27-11./9*M_PI*M_PI-2.*ZETA3)*L
                           +5105./162+67.*M_PI*M_PI/36+M_PI*M_PI*M_PI*M_PI/72-143./9*ZETA3)+CF*T_R*nf*(4.*L-67./3+16.*ZETA3)
                    +CA*T_R*nf*(-4./9*L*L*L+20./9*L*L+(104./27+4.*M_PI*M_PI/9)*L-1832./81-5*M_PI*M_PI/9-92./9*ZETA3))/4/4/M_PI/M_PI;
    CS3 = (CA*pow(nf,2)*(46.59358329522939 + (40*pow(L,3))/81. - (2*pow(L,4))/27. +
                                  pow(L,2)*(1.4320987654320987 + (8*ZETA2)/9.) + (4*ZETA2)/9. +
                                  (4*pow(ZETA2,2))/27. + L*(-19.282578875171467 - (80*ZETA2)/27. -
                                  (128*ZETA3)/27.) + (4576*ZETA3)/243.) +
                                  CF*pow(nf,2)*(55.32098765432099 + (4*pow(L,2))/3. - (20*ZETA2)/9. -
                                  (16*pow(ZETA2,2))/45. - (112*ZETA3)/3. + L*(-17.333333333333332 + (32*ZETA3)/3.))
                                  + pow(CF,2)*nf*(33.77777777777778 - 2*L + (296*ZETA3)/3. - 160*ZETA5) +
                                  pow(CA,2)*nf*(-573.8465172991922 - (8*pow(L,4))/27. + (2*pow(L,5))/9. +
                                  pow(L,3)*(-9.061728395061728 - (2*ZETA2)/3.) - (14189*ZETA2)/729. +
                                  (73*pow(ZETA2,2))/45. - (460*ZETA3)/81. - (82*ZETA2*ZETA3)/9. +
                                  L*(182.49108367626886 + (3820*ZETA2)/81. - (48*pow(ZETA2,2))/5. +
                                  (28*ZETA3)/9.) + pow(L,2)*
                                  (13.962962962962964 - (103*ZETA2)/9. + (118*ZETA3)/9.) + (428*ZETA5)/9.) +
                                  CA*CF*nf*(-351.048353909465 - (8*pow(L,3))/3. - (68*ZETA2)/9. -
                                  (64*pow(ZETA2,2))/45. + L*(70.98148148148148 + 6*ZETA2 +
                                  (16*pow(ZETA2,2))/5. - (376*ZETA3)/9.) + pow(L,2)*(13 - 16*ZETA3) +
                                  (14564*ZETA3)/81. + (64*ZETA2*ZETA3)/3. + (608*ZETA5)/9.) +
                                  pow(CA,3)*(1129.3733043743332 - (11*pow(L,5))/9. - pow(L,6)/6. +
                                  pow(L,4)*(5.203703703703703 - (3*ZETA2)/2.) + (105617*ZETA2)/729. -
                                  (1939*pow(ZETA2,2))/270. - (24389*pow(ZETA2,3))/1890. - (152716*ZETA3)/243. -
                                  (605*ZETA2*ZETA3)/9. - (104*pow(ZETA3,2))/9. +
                                  pow(L,3)*(19.012345679012345 + (11*ZETA2)/3. + 2*ZETA3) +
                                  pow(L,2)*(-83.20987654320987 + (685*ZETA2)/18. - (73*pow(ZETA2,2))/10. +
                                  (143*ZETA3)/9.) + (2222*ZETA5)/9. +
                                  L*(-256.4986282578875 - (13420*ZETA2)/81. + (176*pow(ZETA2,2))/5. +
                                  (2048*ZETA3)/27. + (34*ZETA2*ZETA3)/3. + 16*ZETA5)))/4./4./4./M_PI/M_PI/M_PI;
     return;
  }
  */



};






